/*
 * This sketch is controling a RGB123 board (16x16) with a RFDuino Arduino module 
 * This last is a Arm-Cortex M0 @ 16Mhz
 * Due to really short timing of WS2812b leds running under 1us the update procedure has to be 
 * written in assembly language and optimized for this purpose
 * 
 * ** This sketch only works with this specific Arduino board (or eventually Cortex-m0 @ 16Mhz others) **
 *
 *
 * Copyright (c) 2014 Paul Pinault www.disk91.com .  All right reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * Author : Paul Pinault - www.disk91.com
 *          follow me on twitter : @disk_91
 */

#include <RFduinoBLE.h>

// GPIO2 heartbit
int led1 = 6;
// GPIO1
int toRGB = 2;


void setup() {
  Serial.begin(9600);
  
  // setup the heartbit leds for output
  pinMode(led1, OUTPUT);
  rgb123_256_init(toRGB,16,16);
}

void loop() {
  int v = 0;
  int w = 0;
  int z = 0;
  Serial.println("startup !");

  // set max Intensity to 2500 mA
  rgb123_256_setIntensityLimit(2500);

  int l=0;
  while (1) {
    int x,y;
     for ( y = 0 ; y < 16 ; y++ ) {
       for ( x = 0 ; x < 16 ; x++ ) {
         rgb123_256_setPixel(x, y, 0xff, 0xff, 0xff);
         if ( x+y*16 > l ) break;
       }
       if ( x+y*16 > l ) break;
     }
     Serial.print("l :");Serial.print(l);Serial.print(" mA:");Serial.println(rgb123_256_estimateCurrent());
     rgb123_256_showPixels(toRGB);
     Serial.print("l :");Serial.print(l);Serial.print(" mA - corrected:");Serial.println(rgb123_256_estimateCurrent());
     delay(500);
     l = (l + 1) & 0xff;
     rgb123_256_clrPixels();
  }
 

}


/* ==========================================================================
 * RGB123 functions
 * ==========================================================================
 */

// Maximum supported LED (this can be changed) 
#define MAX_lEDS  256

// ---------------------------------------------------
// Raw data representing pixels each value format is
// --------bbbbbbbbrrrrrrrrgggggggg
// --------012345670123456701234567
// led matrix is 16x16, direction change each line
uint32_t rgb123_256_pixels[MAX_lEDS];
int max_x=0;  // number of Leds on a line
int max_y=0;  // number of Leds

int max_i=-1; // maximum intensity in mA => -1 = no limit


// --------------------------------------------------
// Initialize rgb123 panel and internal memory
// structure.
// @pin : control pin used to communicate with panel
// @mx  : number of leds per lines 
// @my  : number of lines 
// Rq: mx*my must be < MAX_LEDS
void rgb123_256_init( uint32_t pin, int mx, int my ) {
   pinMode(pin, OUTPUT);
   // 50 us low voltage for reset
   NRF_GPIO->OUTCLR = ( 1 << pin );
   // set max X and Y
   if ( mx * my <= MAX_lEDS ) {
     max_x=mx;
     max_y=my;
   } else {
     max_x=1;
     max_y=1;
   }
} 

// ---------------------------------------------------
// set a pixel in the coordinate [0-15][0-15]
void rgb123_256_setPixel(uint32_t x, uint32_t y, uint32_t r, uint32_t g, uint32_t b) {
  if ( x >= max_x || y >= max_y ) return;
  
  int v = 0;
  for ( int i = 0 ; i < 8 ; i++ ) 
    v = ( v | (( b >> i ) & 1) ) << 1;
  for ( int i = 0 ; i < 8 ; i++ ) 
    v = ( v | (( r >> i ) & 1) ) << 1;
  for ( int i = 0 ; i < 8 ; i++ ) 
    v = ( v | (( g >> i ) & 1) ) << 1;
  v = v >> 1;
  
  uint32_t _x = (y & 1 != 0)?(max_x-1)-x:x;
  rgb123_256_pixels[max_x*y+_x] = v;
}

// --------------------------------------------------
// get r,g,b values from the matrix
// @TODO - check it works
void rgb123_256_getPixel(uint32_t x, uint32_t y, uint32_t * r, uint32_t * g, uint32_t * b) {
    if ( x >= max_x || y >= max_y ) return;
    uint32_t _x = (y & 1 != 0)?(max_x-1)-x:x;
    int v = rgb123_256_pixels[max_x*y+_x];
    
    *g=0;
    *r=0;
    *b=0;
    for ( int i = 0 ; i < 8 ; i++ )
       *g = ( *g | (( v >> i ) & 1 )) << 1;
    *g = *g >> 1;
    
    for ( int i = 8 ; i < 16 ; i++ )
       *r = ( *r | (( v >> i ) & 1 )) << 1;
    *r = *r >> 1;
    
    for ( int i = 16 ; i < 24 ; i++ )
       *b = ( *b | (( v >> i ) & 1 )) << 1;
    *b = *b >> 1; 
}
  
// --------------------------------------------------
// evaluate current consumption of the 16x16 matrix
// per pixel : ma = 1 + R/16 + G/16 + B/16
// @TODO - finish it !
int rgb123_256_estimateCurrent() {
  int ma = 0;
  uint32_t r,g,b;
  for ( int y = 0 ; y < max_y ; y++ ) {
    for ( int x = 0 ; x < max_x ; x++ ) {
       rgb123_256_getPixel(x,y,&r,&g,&b);
       ma += 1 + (r >> 4) + ( g >> 4) + ( b >> 4 );      
    }
  }
  return ma;
}

// --------------------------------------------------
// setIntensityLimit - 5% error
void rgb123_256_setIntensityLimit(int _m) {
  max_i=(_m - _m/20 );
  rgb123_256_limitIntensity();
}

// --------------------------------------------------
// setIntensityLimit
void rgb123_256_limitIntensity() {
  int _initialI = rgb123_256_estimateCurrent();
  if ( _max_i > 0 && _initialI > max_i ) {
     // ratio computation, the fixed part can't be reduce and is substracted from max_i
     int ratio = ( (max_i - max_x*max_y) << 10 ) / (_initialI - max_x*max_y);
     
     uint32_t r,g,b;
     for ( int y = 0 ; y < max_y ; y++ ) {
       for ( int x = 0 ; x < max_x ; x++ ) {
          rgb123_256_getPixel(x,y,&r,&g,&b);
          r = (( r * ratio ) >> 10);
          g = (( g * ratio ) >> 10);
          b = (( b * ratio ) >> 10);
          rgb123_256_setPixel(x,y,r,g,b); 
       }
     }
    
  }
}



// ---------------------------------------------------
// clear the pixel matrix
void rgb123_256_clrPixels() {
  for (int i = 0 ; i < MAX_lEDS ; i++) rgb123_256_pixels[i] = 0;
}


// ---------------------------------------------------
// display pixel matrix
// max = 1 << 24                                    
// end = _ptbits + 256 * 4 + 4                      
// while ( _ptbits != end ) {      
//   v = *_ptbits                                   
//      b = v & 1
//      if (b) write(1) else write(0) ; v = v << 1
//      if (b) write(1) else write(0) ; v = v << 1
//      ... x24 ...
//      if (b) write(1) else write(0) ; v = v << 1
//   _ptbits++;
// }
void rgb123_256_showPixels(uint32_t pin) {
  
  rgb123_256_limitIntensity();
  
  volatile uint32_t * outset = &NRF_GPIO->OUTSET;
  volatile uint32_t * outclr = &NRF_GPIO->OUTCLR;
  uint32_t * _ptbits = rgb123_256_pixels;
  uint32_t * _end    = &rgb123_256_pixels[max_x*max_y];   
  uint32_t   _max    = ( 1 << 24 );
  uint32_t   _p      = 1 << pin;


  noInterrupts();
  
  // outset  = %0
  // outclr  = %1
  // _ptbits = %2
  // _end    = %3
  // _max    = %4
  // _p      = %5
  //
  // v       = r7
  // i       = r6
  // b       = r5
  asm volatile (
      "headSP:\n\t"
      "push {r0,r1,r2,r3,r4,r5,r6,r7}\n\t"
      "wh01: \n\t"                // while ( _ptbits < _end ) {
         "cmp %2,%3 \n\t"         //    1 - cycle
         "blo wh01s \n\t"         //    3 - cycles
         "b   ewh01 \n\t"

       "wh01s: \n\t"  
         "ldr r7,[%2,#0] \n\t"    //    v = *ptbits
         "mov r6,#1 \n\t"         //    if ( v & 1 == 0 )
         "and r6,r7 \n\t"      
       "wh01_optim: \n\t"          // optimized loop return from end
         "str %5,[%0,#0] \n\t"  //    *outset = _p 
         "cmp r6,#0 \n\t"      
         "beq b0zero \n\t"
         "b0one: \n\t"            // case 1 ( 11-15 cycle high then 5 - 9 cycle low )
           "lsr  r7,r7,#1 \n\t"   //    -- cycle 3 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
           "nop \n\t"             //    -- cycle 11
           "str %5,[%1,#0] \n\t"  //    -- cycle 12-13
           "b b0end \n\t"         //    -- cycle 1-2-3
         "b0zero: \n\t"           // case 0 ( 4-8 cycles high then 12-16 cycles low )       
           "str %5,[%1,#0] \n\t"  //    -- cycle 5-6
           "lsr r7,r7,#1 \n\t"   //    -- cycle 1 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 2
           "nop \n\t"             //    -- cycle 3
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
        "b0end: \n\t"             //    -- 4 more cycle before high
        
         "mov r6,#1 \n\t"         //    if ( v & 1 == 0 )
         "and r6,r7 \n\t"
         "str %5,[%0,#0] \n\t"  //    *outset = _p
         "cmp r6,#0 \n\t"      
         "beq b1zero \n\t"
         "b1one: \n\t"            // case 1 ( 11-15 cycle high then 5 - 9 cycle low )
           "lsr r7,r7,#1 \n\t"   //    -- cycle 3 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
           "nop \n\t"             //    -- cycle 11
           "str %5,[%1,#0] \n\t"  //    -- cycle 12-13
           "b b1end \n\t"         //    -- cycle 1-2-3
         "b1zero: \n\t"           // case 0 ( 4-8 cycles high then 12-16 cycles low )       
           "str %5,[%1,#0] \n\t"  //    -- cycle 5-6
           "lsr r7,r7,#1 \n\t"   //    -- cycle 1 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 2
           "nop \n\t"             //    -- cycle 3
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
        "b1end: \n\t"             //    -- 4 more cycle before high
           

         "mov r6,#1 \n\t"         //    if ( v & 1 == 0 )
         "and r6,r7 \n\t"
         "str %5,[%0,#0] \n\t"  //    *outset = _p
         "cmp r6,#0 \n\t"      
         "beq b2zero \n\t"
         "b2one: \n\t"            // case 1 ( 11-15 cycle high then 5 - 9 cycle low )
           "lsr r7,r7,#1 \n\t"   //    -- cycle 3 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
           "nop \n\t"             //    -- cycle 11
           "str %5,[%1,#0] \n\t"  //    -- cycle 12-13
           "b b2end \n\t"         //    -- cycle 1-2-3
         "b2zero: \n\t"           // case 0 ( 4-8 cycles high then 12-16 cycles low )       
           "str %5,[%1,#0] \n\t"  //    -- cycle 5-6
           "lsr r7,r7,#1 \n\t"   //    -- cycle 1 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 2
           "nop \n\t"             //    -- cycle 3
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
        "b2end: \n\t"             //    -- 4 more cycle before high

         "mov r6,#1 \n\t"         //    if ( v & 1 == 0 )
         "and r6,r7 \n\t"
         "str %5,[%0,#0] \n\t"  //    *outset = _p
         "cmp r6,#0 \n\t"      
         "beq b3zero \n\t"
         "b3one: \n\t"            // case 1 ( 11-15 cycle high then 5 - 9 cycle low )
           "lsr r7,r7,#1 \n\t"   //    -- cycle 3 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
           "nop \n\t"             //    -- cycle 11
           "str %5,[%1,#0] \n\t"  //    -- cycle 12-13
           "b b3end \n\t"         //    -- cycle 1-2-3
         "b3zero: \n\t"           // case 0 ( 4-8 cycles high then 12-16 cycles low )       
           "str %5,[%1,#0] \n\t"  //    -- cycle 5-6
           "lsr r7,r7,#1 \n\t"   //    -- cycle 1 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 2
           "nop \n\t"             //    -- cycle 3
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
        "b3end: \n\t"             //    -- 4 more cycle before high

         "mov r6,#1 \n\t"         //    if ( v & 1 == 0 )
         "and r6,r7 \n\t"
         "str %5,[%0,#0] \n\t"  //    *outset = _p
         "cmp r6,#0 \n\t"      
         "beq b4zero \n\t"
         "b4one: \n\t"            // case 1 ( 11-15 cycle high then 5 - 9 cycle low )
           "lsr r7,r7,#1 \n\t"   //    -- cycle 3 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
           "nop \n\t"             //    -- cycle 11
           "str %5,[%1,#0] \n\t"  //    -- cycle 12-13
           "b b4end \n\t"         //    -- cycle 1-2-3
         "b4zero: \n\t"           // case 0 ( 4-8 cycles high then 12-16 cycles low )       
           "str %5,[%1,#0] \n\t"  //    -- cycle 5-6
           "lsr r7,r7,#1 \n\t"   //    -- cycle 1 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 2
           "nop \n\t"             //    -- cycle 3
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
        "b4end: \n\t"             //    -- 4 more cycle before high

         "mov r6,#1 \n\t"         //    if ( v & 1 == 0 )
         "and r6,r7 \n\t"
         "str %5,[%0,#0] \n\t"  //    *outset = _p
         "cmp r6,#0 \n\t"      
         "beq b5zero \n\t"
         "b5one: \n\t"            // case 1 ( 11-15 cycle high then 5 - 9 cycle low )
           "lsr r7,r7,#1 \n\t"   //    -- cycle 3 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
           "nop \n\t"             //    -- cycle 11
           "str %5,[%1,#0] \n\t"  //    -- cycle 12-13
           "b b5end \n\t"         //    -- cycle 1-2-3
         "b5zero: \n\t"           // case 0 ( 4-8 cycles high then 12-16 cycles low )       
           "str %5,[%1,#0] \n\t"  //    -- cycle 5-6
           "lsr r7,r7,#1 \n\t"   //    -- cycle 1 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 2
           "nop \n\t"             //    -- cycle 3
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
        "b5end: \n\t"             //    -- 4 more cycle before high

         "mov r6,#1 \n\t"         //    if ( v & 1 == 0 )
         "and r6,r7 \n\t"
         "str %5,[%0,#0] \n\t"  //    *outset = _p
         "cmp r6,#0 \n\t"      
         "beq b6zero \n\t"
         "b6one: \n\t"            // case 1 ( 11-15 cycle high then 5 - 9 cycle low )
           "lsr r7,r7,#1 \n\t"   //    -- cycle 3 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
           "nop \n\t"             //    -- cycle 11
           "str %5,[%1,#0] \n\t"  //    -- cycle 12-13
           "b b6end \n\t"         //    -- cycle 1-2-3
         "b6zero: \n\t"           // case 0 ( 4-8 cycles high then 12-16 cycles low )       
           "str %5,[%1,#0] \n\t"  //    -- cycle 5-6
           "lsr r7,r7,#1 \n\t"   //    -- cycle 1 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 2
           "nop \n\t"             //    -- cycle 3
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
        "b6end: \n\t"             //    -- 4 more cycle before high

         "mov r6,#1 \n\t"         //    if ( v & 1 == 0 )
         "and r6,r7 \n\t"
         "str %5,[%0,#0] \n\t"  //    *outset = _p
         "cmp r6,#0 \n\t"      
         "beq b7zero \n\t"
         "b7one: \n\t"            // case 1 ( 11-15 cycle high then 5 - 9 cycle low )
           "lsr r7,r7,#1 \n\t"   //    -- cycle 3 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
           "nop \n\t"             //    -- cycle 11
           "str %5,[%1,#0] \n\t"  //    -- cycle 12-13
           "b b7end \n\t"         //    -- cycle 1-2-3
         "b7zero: \n\t"           // case 0 ( 4-8 cycles high then 12-16 cycles low )       
           "str %5,[%1,#0] \n\t"  //    -- cycle 5-6
           "lsr r7,r7,#1 \n\t"   //    -- cycle 1 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 2
           "nop \n\t"             //    -- cycle 3
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
        "b7end: \n\t"             //    -- 4 more cycle before high

         "mov r6,#1 \n\t"         //    if ( v & 1 == 0 )
         "and r6,r7 \n\t"
         "str %5,[%0,#0] \n\t"  //    *outset = _p
         "cmp r6,#0 \n\t"      
         "beq b8zero \n\t"
         "b8one: \n\t"            // case 1 ( 11-15 cycle high then 5 - 9 cycle low )
           "lsr r7,r7,#1 \n\t"   //    -- cycle 3 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
           "nop \n\t"             //    -- cycle 11
           "str %5,[%1,#0] \n\t"  //    -- cycle 12-13
           "b b8end \n\t"         //    -- cycle 1-2-3
         "b8zero: \n\t"           // case 0 ( 4-8 cycles high then 12-16 cycles low )       
           "str %5,[%1,#0] \n\t"  //    -- cycle 5-6
           "lsr r7,r7,#1 \n\t"   //    -- cycle 1 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 2
           "nop \n\t"             //    -- cycle 3
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
        "b8end: \n\t"             //    -- 4 more cycle before high

         "mov r6,#1 \n\t"         //    if ( v & 1 == 0 )
         "and r6,r7 \n\t"
         "str %5,[%0,#0] \n\t"  //    *outset = _p
         "cmp r6,#0 \n\t"      
         "beq b9zero \n\t"
         "b9one: \n\t"            // case 1 ( 11-15 cycle high then 5 - 9 cycle low )
           "lsr r7,r7,#1 \n\t"   //    -- cycle 3 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
           "nop \n\t"             //    -- cycle 11
           "str %5,[%1,#0] \n\t"  //    -- cycle 12-13
           "b b9end \n\t"         //    -- cycle 1-2-3
         "b9zero: \n\t"           // case 0 ( 4-8 cycles high then 12-16 cycles low )       
           "str %5,[%1,#0] \n\t"  //    -- cycle 5-6
           "lsr r7,r7,#1 \n\t"   //    -- cycle 1 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 2
           "nop \n\t"             //    -- cycle 3
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
        "b9end: \n\t"             //    -- 4 more cycle before high

         "mov r6,#1 \n\t"         //    if ( v & 1 == 0 )
         "and r6,r7 \n\t"
         "str %5,[%0,#0] \n\t"  //    *outset = _p
         "cmp r6,#0 \n\t"      
         "beq b10zero \n\t"
         "b10one: \n\t"            // case 1 ( 11-15 cycle high then 5 - 9 cycle low )
           "lsr r7,r7,#1 \n\t"   //    -- cycle 3 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
           "nop \n\t"             //    -- cycle 11
           "str %5,[%1,#0] \n\t"  //    -- cycle 12-13
           "b b10end \n\t"         //    -- cycle 1-2-3
         "b10zero: \n\t"           // case 0 ( 4-8 cycles high then 12-16 cycles low )       
           "str %5,[%1,#0] \n\t"  //    -- cycle 5-6
           "lsr r7,r7,#1 \n\t"   //    -- cycle 1 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 2
           "nop \n\t"             //    -- cycle 3
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
        "b10end: \n\t"             //    -- 4 more cycle before high

         "mov r6,#1 \n\t"         //    if ( v & 1 == 0 )
         "and r6,r7 \n\t"
         "str %5,[%0,#0] \n\t"  //    *outset = _p
         "cmp r6,#0 \n\t"      
         "beq b11zero \n\t"
         "b11one: \n\t"            // case 1 ( 11-15 cycle high then 5 - 9 cycle low )
           "lsr r7,r7,#1 \n\t"   //    -- cycle 3 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
           "nop \n\t"             //    -- cycle 11
           "str %5,[%1,#0] \n\t"  //    -- cycle 12-13
           "b b11end \n\t"         //    -- cycle 1-2-3
         "b11zero: \n\t"           // case 0 ( 4-8 cycles high then 12-16 cycles low )       
           "str %5,[%1,#0] \n\t"  //    -- cycle 5-6
           "lsr r7,r7,#1 \n\t"   //    -- cycle 1 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 2
           "nop \n\t"             //    -- cycle 3
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
        "b11end: \n\t"             //    -- 4 more cycle before high

         "mov r6,#1 \n\t"         //    if ( v & 1 == 0 )
         "and r6,r7 \n\t"
         "str %5,[%0,#0] \n\t"  //    *outset = _p
         "cmp r6,#0 \n\t"      
         "beq b12zero \n\t"
         "b12one: \n\t"            // case 1 ( 11-15 cycle high then 5 - 9 cycle low )
           "lsr r7,r7,#1 \n\t"   //    -- cycle 3 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
           "nop \n\t"             //    -- cycle 11
           "str %5,[%1,#0] \n\t"  //    -- cycle 12-13
           "b b12end \n\t"         //    -- cycle 1-2-3
         "b12zero: \n\t"           // case 0 ( 4-8 cycles high then 12-16 cycles low )       
           "str %5,[%1,#0] \n\t"  //    -- cycle 5-6
           "lsr r7,r7,#1 \n\t"   //    -- cycle 1 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 2
           "nop \n\t"             //    -- cycle 3
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
        "b12end: \n\t"             //    -- 4 more cycle before high

         "mov r6,#1 \n\t"         //    if ( v & 1 == 0 )
         "and r6,r7 \n\t"
         "str %5,[%0,#0] \n\t"  //    *outset = _p
         "cmp r6,#0 \n\t"      
         "beq b13zero \n\t"
         "b13one: \n\t"            // case 1 ( 11-15 cycle high then 5 - 9 cycle low )
           "lsr r7,r7,#1 \n\t"   //    -- cycle 3 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
           "nop \n\t"             //    -- cycle 11
           "str %5,[%1,#0] \n\t"  //    -- cycle 12-13
           "b b13end \n\t"         //    -- cycle 1-2-3
         "b13zero: \n\t"           // case 0 ( 4-8 cycles high then 12-16 cycles low )       
           "str %5,[%1,#0] \n\t"  //    -- cycle 5-6
           "lsr r7,r7,#1 \n\t"   //    -- cycle 1 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 2
           "nop \n\t"             //    -- cycle 3
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
        "b13end: \n\t"             //    -- 4 more cycle before high

         "mov r6,#1 \n\t"         //    if ( v & 1 == 0 )
         "and r6,r7 \n\t"
         "str %5,[%0,#0] \n\t"  //    *outset = _p
         "cmp r6,#0 \n\t"      
         "beq b14zero \n\t"
         "b14one: \n\t"            // case 1 ( 11-15 cycle high then 5 - 9 cycle low )
           "lsr r7,r7,#1 \n\t"   //    -- cycle 3 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
           "nop \n\t"             //    -- cycle 11
           "str %5,[%1,#0] \n\t"  //    -- cycle 12-13
           "b b14end \n\t"         //    -- cycle 1-2-3
         "b14zero: \n\t"           // case 0 ( 4-8 cycles high then 12-16 cycles low )       
           "str %5,[%1,#0] \n\t"  //    -- cycle 5-6
           "lsr r7,r7,#1 \n\t"   //    -- cycle 1 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 2
           "nop \n\t"             //    -- cycle 3
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
        "b14end: \n\t"             //    -- 4 more cycle before high

         "mov r6,#1 \n\t"         //    if ( v & 1 == 0 )
         "and r6,r7 \n\t"
         "str %5,[%0,#0] \n\t"  //    *outset = _p
         "cmp r6,#0 \n\t"      
         "beq b15zero \n\t"
         "b15one: \n\t"            // case 1 ( 11-15 cycle high then 5 - 9 cycle low )
           "lsr r7,r7,#1 \n\t"   //    -- cycle 3 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
           "nop \n\t"             //    -- cycle 11
           "str %5,[%1,#0] \n\t"  //    -- cycle 12-13
           "b b15end \n\t"         //    -- cycle 1-2-3
         "b15zero: \n\t"           // case 0 ( 4-8 cycles high then 12-16 cycles low )       
           "str %5,[%1,#0] \n\t"  //    -- cycle 5-6
           "lsr r7,r7,#1 \n\t"   //    -- cycle 1 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 2
           "nop \n\t"             //    -- cycle 3
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
        "b15end: \n\t"             //    -- 4 more cycle before high

         "mov r6,#1 \n\t"         //    if ( v & 1 == 0 )
         "and r6,r7 \n\t"
         "str %5,[%0,#0] \n\t"  //    *outset = _p
         "cmp r6,#0 \n\t"      
         "beq b16zero \n\t"
         "b16one: \n\t"            // case 1 ( 11-15 cycle high then 5 - 9 cycle low )
           "lsr r7,r7,#1 \n\t"   //    -- cycle 3 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
           "nop \n\t"             //    -- cycle 11
           "str %5,[%1,#0] \n\t"  //    -- cycle 12-13
           "b b16end \n\t"         //    -- cycle 1-2-3
         "b16zero: \n\t"           // case 0 ( 4-8 cycles high then 12-16 cycles low )       
           "str %5,[%1,#0] \n\t"  //    -- cycle 5-6
           "lsr r7,r7,#1 \n\t"   //    -- cycle 1 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 2
           "nop \n\t"             //    -- cycle 3
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
        "b16end: \n\t"             //    -- 4 more cycle before high

         "mov r6,#1 \n\t"         //    if ( v & 1 == 0 )
         "and r6,r7 \n\t"
         "str %5,[%0,#0] \n\t"  //    *outset = _p
         "cmp r6,#0 \n\t"      
         "beq b17zero \n\t"
         "b17one: \n\t"            // case 1 ( 11-15 cycle high then 5 - 9 cycle low )
           "lsr r7,r7,#1 \n\t"   //    -- cycle 3 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
           "nop \n\t"             //    -- cycle 11
           "str %5,[%1,#0] \n\t"  //    -- cycle 12-13
           "b b17end \n\t"         //    -- cycle 1-2-3
         "b17zero: \n\t"           // case 0 ( 4-8 cycles high then 12-16 cycles low )       
           "str %5,[%1,#0] \n\t"  //    -- cycle 5-6
           "lsr r7,r7,#1 \n\t"   //    -- cycle 1 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 2
           "nop \n\t"             //    -- cycle 3
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
        "b17end: \n\t"             //    -- 4 more cycle before high

         "mov r6,#1 \n\t"         //    if ( v & 1 == 0 )
         "and r6,r7 \n\t"
         "str %5,[%0,#0] \n\t"  //    *outset = _p
         "cmp r6,#0 \n\t"      
         "beq b18zero \n\t"
         "b18one: \n\t"            // case 1 ( 11-15 cycle high then 5 - 9 cycle low )
           "lsr r7,r7,#1 \n\t"   //    -- cycle 3 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
           "nop \n\t"             //    -- cycle 11
           "str %5,[%1,#0] \n\t"  //    -- cycle 12-13
           "b b18end \n\t"         //    -- cycle 1-2-3
         "b18zero: \n\t"           // case 0 ( 4-8 cycles high then 12-16 cycles low )       
           "str %5,[%1,#0] \n\t"  //    -- cycle 5-6
           "lsr r7,r7,#1 \n\t"   //    -- cycle 1 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 2
           "nop \n\t"             //    -- cycle 3
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
        "b18end: \n\t"             //    -- 4 more cycle before high

         "mov r6,#1 \n\t"         //    if ( v & 1 == 0 )
         "and r6,r7 \n\t"
         "str %5,[%0,#0] \n\t"  //    *outset = _p
         "cmp r6,#0 \n\t"      
         "beq b19zero \n\t"
         "b19one: \n\t"            // case 1 ( 11-15 cycle high then 5 - 9 cycle low )
           "lsr r7,r7,#1 \n\t"   //    -- cycle 3 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
           "nop \n\t"             //    -- cycle 11
           "str %5,[%1,#0] \n\t"  //    -- cycle 12-13
           "b b19end \n\t"         //    -- cycle 1-2-3
         "b19zero: \n\t"           // case 0 ( 4-8 cycles high then 12-16 cycles low )       
           "str %5,[%1,#0] \n\t"  //    -- cycle 5-6
           "lsr r7,r7,#1 \n\t"   //    -- cycle 1 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 2
           "nop \n\t"             //    -- cycle 3
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
        "b19end: \n\t"             //    -- 4 more cycle before high

         "mov r6,#1 \n\t"         //    if ( v & 1 == 0 )
         "and r6,r7 \n\t"
         "str %5,[%0,#0] \n\t"  //    *outset = _p
         "cmp r6,#0 \n\t"      
         "beq b20zero \n\t"
         "b20one: \n\t"            // case 1 ( 11-15 cycle high then 5 - 9 cycle low )
           "lsr r7,r7,#1 \n\t"   //    -- cycle 3 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
           "nop \n\t"             //    -- cycle 11
           "str %5,[%1,#0] \n\t"  //    -- cycle 12-13
           "b b20end \n\t"         //    -- cycle 1-2-3
         "b20zero: \n\t"           // case 0 ( 4-8 cycles high then 12-16 cycles low )       
           "str %5,[%1,#0] \n\t"  //    -- cycle 5-6
           "lsr r7,r7,#1 \n\t"   //    -- cycle 1 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 2
           "nop \n\t"             //    -- cycle 3
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
        "b20end: \n\t"             //    -- 4 more cycle before high

         "mov r6,#1 \n\t"         //    if ( v & 1 == 0 )
         "and r6,r7 \n\t"
         "str %5,[%0,#0] \n\t"  //    *outset = _p
         "cmp r6,#0 \n\t"      
         "beq b21zero \n\t"
         "b21one: \n\t"            // case 1 ( 11-15 cycle high then 5 - 9 cycle low )
           "lsr r7,r7,#1 \n\t"   //    -- cycle 3 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
           "nop \n\t"             //    -- cycle 11
           "str %5,[%1,#0] \n\t"  //    -- cycle 12-13
           "b b21end \n\t"         //    -- cycle 1-2-3
         "b21zero: \n\t"           // case 0 ( 4-8 cycles high then 12-16 cycles low )       
           "str %5,[%1,#0] \n\t"  //    -- cycle 5-6
           "lsr r7,r7,#1 \n\t"   //    -- cycle 1 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 2
           "nop \n\t"             //    -- cycle 3
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
        "b21end: \n\t"             //    -- 4 more cycle before high

         "mov r6,#1 \n\t"         //    if ( v & 1 == 0 )
         "and r6,r7 \n\t"
         "str %5,[%0,#0] \n\t"  //    *outset = _p
         "cmp r6,#0 \n\t"      
         "beq b22zero \n\t"
         "b22one: \n\t"            // case 1 ( 11-15 cycle high then 5 - 9 cycle low )
           "lsr r7,r7,#1 \n\t"   //    -- cycle 3 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
           "nop \n\t"             //    -- cycle 11
           "str %5,[%1,#0] \n\t"  //    -- cycle 12-13
           "b b22end \n\t"         //    -- cycle 1-2-3
         "b22zero: \n\t"           // case 0 ( 4-8 cycles high then 12-16 cycles low )       
           "str %5,[%1,#0] \n\t"  //    -- cycle 5-6
           "lsr r7,r7,#1 \n\t"   //    -- cycle 1 anticipate next bit extract ( v = v >> 1)
           "nop \n\t"             //    -- cycle 2
           "nop \n\t"             //    -- cycle 3
           "nop \n\t"             //    -- cycle 4
           "nop \n\t"             //    -- cycle 5
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
        "b22end: \n\t"             //    -- 4 more cycle before high

         "mov r6,#1 \n\t"         //    if ( v & 1 == 0 )
         "and r6,r7 \n\t"
         "str %5,[%0,#0] \n\t"  //    *outset = _p
         "cmp r6,#0 \n\t"      
         "beq b23zero \n\t"
         "b23one: \n\t"            // case 1 ( 11-15 cycle high then 5 - 9 cycle low )
           "add %2,%2,#4 \n\t"    //    -- cycle 3 - anticipation _ptbits++
           "ldr r7,[%2,#0] \n\t"  //    -- cycle 4-5 - anticipation v = *ptbits
           "mov r6,#1 \n\t"       //    -- cycle 6 - anticipation if ( v & 1 == 0 )
           "and r6,r7 \n\t"       //    -- cycle 7 - anticipation
           "nop \n\t"             //    -- cycle 8
           "nop \n\t"             //    -- cycle 9
           "nop \n\t"             //    -- cycle 10
           "nop \n\t"             //    -- cycle 11
           "str %5,[%1,#0] \n\t"  //    -- cycle 12-13
           "cmp %2,%3 \n\t"       //    -- cycle 1 - check end of wh01 loop
           "bhs ewh01 \n\t"       //    -- cycle 2 - if yes goes out
           "b wh01_optim \n\t"    //    -- cycle 3-4-5 + 2 before signal goes up
         "b23zero: \n\t"           // case 0 ( 4-8 cycles high then 12-16 cycles low )       
           "str %5,[%1,#0] \n\t"  //    -- cycle 5-6
           "add %2,%2,#4 \n\t"    //    -- cycle 1 - anticipation _ptbits++
           "ldr r7,[%2,#0] \n\t"  //    -- cycle 2-3 - anticipation v = *ptbits
           "mov r6,#1 \n\t"       //    -- cycle 4 - anticipation if ( v & 1 == 0 )
           "and r6,r7 \n\t"       //    -- cycle 5 - anticipation
           "nop \n\t"             //    -- cycle 6
           "nop \n\t"             //    -- cycle 7
           "cmp %2,%3 \n\t"       //    -- cycle 8 - check end of wh01 loop
           "bhs ewh01 \n\t"       //    -- cycle 9 - if yes goes out
           "b wh01_optim \n\t"    //    -- cycle 10-11-12 + 2 before signal goes up

       "ewh01: \n\t" 
       "pop {r0,r1,r2,r3,r4,r5,r6,r7}\n\t" 
    ::
    "r" (outset),   // %0
    "r" (outclr),   // %1
    "r" (_ptbits),  // %2
    "r" (_end),     // %3
    "r" (_max),     // %4
    "r" (_p)        // %5
 );
  
    interrupts();
    delayMicroseconds(60);
}

